"""Structure model elements.

:ref: https://typst.app/docs/reference/model/
"""
# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107

from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

import jinja2

from .base import Element

if TYPE_CHECKING:
    from typing import Optional


class Document(Element):
    """Document element."""

    LABEL = "document"
    TEMPLATE = """\
        {% for content in contents %}
        {{ content }}
        {% endfor %}
    """


class Section(Element):
    """Sphinx's section element."""

    LABEL = "section"
    TEMPLATE: str = """\
        {%- for content in contents -%}
        {{ content }}
        {% endfor %}
    """


class Heading(Element):
    """Section's heading element.

    :ref: https://typst.app/docs/reference/model/heading/
    """

    LABEL = "heading"
    TEMPLATE = """\
        #heading(
          level: {{level}},
          [
            {{content}}
            {%- if label %}
            #label("{{ label }}")
            {%- endif %}
          ]
        )
    """

    level: int = 0
    label: Optional[str] = None
    """RefID of document."""

    def to_text(self):
        content = self.children[0].to_text() if self.children else ""
        if self.level > 0:
            return self.get_template().render(
                level=self.level, content=content, label=self.label
            )
        return ""


class Paragraph(Element):
    """Paragraph of document."""

    LABEL = "par"
    TEMPLATE: str = """\
        {%- for content in contents -%}
        {{ content }}
        {%- endfor %}
    """


class List(Element):
    """Abstract element of list-type content."""

    TEMPLATE = """\
        {{ prefix }}{{ funcname }}(
          {%- for content, delimiter in contents %}
          {%- if not loop.first %}{{delimiter}}{% endif %}
          {{ content | indent(2, first=False) }}
          {%- endfor %}
        )
    """

    def to_text(self):
        """Render as Typst source.

        Typst must remove function symbol to use nested list.
        It replace ``+`` to join for parent.
        """
        prefix = "+" if isinstance(self.parent, List) else "#"
        contents = []
        for c in self.children:
            if isinstance(c, List):
                text = c.to_text()
                delimiter = ""
            else:
                text = jinja2.Template(
                    textwrap.dedent("""\
                    [
                      {{c|indent(2, first=False)}}
                    ]
                """).rstrip("\n")
                ).render(c=c.to_text())
                delimiter = ","
            contents.append((text, delimiter))
        return self.get_template().render(
            prefix=prefix, funcname=self.LABEL, contents=contents
        )


class BulletList(List):
    """Bullet list element.

    :ref: https://typst.app/docs/reference/model/list/
    """

    LABEL = "list"


class NumberedList(List):
    """Enumerated list element.

    :ref: https://typst.app/docs/reference/model/enum/
    """

    LABEL = "enum"


class Table(Element):
    """Table content element.

    .. note:: Currently, some elements use this if it is not table directive.

    .. todo:: This only renders simple style table, It should support thead design.

    :ref: https://typst.app/docs/reference/model/table/
    """

    LABEL = "table"
    TEMPLATE = """\
        #table(
          columns: {{ columns }},
          {%- for content in contents %}
          {%- if not loop.first %},{% endif %}
          [
            {{ content | indent(4, first=False) }}
          ]
          {%- endfor %}
        )
    """

    columns: int = 2

    def to_text(self):
        contents = [f"{c.to_text()}" for c in self.children]
        return self.get_template().render(contents=contents, columns=self.columns)


class Quote(Element):
    """Blockquote element.

    :ref: https://typst.app/docs/reference/model/quote/
    """

    LABEL = "quote"
    TEMPLATE = """\
        #quote(
          block: true,
          {%- if attribution %}
          attribution: [{{attribution}}],
          {%- endif %}
        )[
          {%- for content in contents %}
          {{ content | indent(2, first=False) }}
          {%- endfor %}
        ]
    """

    attribution: str = ""

    def to_text(self):
        return self.get_template().render(
            contents=[c.to_text() for c in self.children],
            attribution=self.attribution,
        )


class Field(Element):
    LABEL = "fiield"
    TEMPLATE = """\
        #sphinxField(
          [
            {{ title|indent(4, first=False) }}
          ],
          [
            {%- for content in contents %}
            {{ content | indent(4, first=False) }}
            {%- endfor %}
          ],
        )
    """

    def to_text(self):
        return self.get_template().render(
            title=self.children[0].to_text(),
            contents=[c.to_text() for c in self.children[1:]],
        )


class Figure(Element):
    """Component element included image and caption.

    :ref: https://typst.app/docs/reference/model/figure/
    """

    LABEL = "figure"
    TEMPLATE = """\
        #figure(
          {{ image|indent(2, first=False) }},
          {%- if caption %}
          caption: [
            {%- for content in caption %}
            {{ content | indent(4, first=False) }}
            {%- endfor %}
          ],
          {%- endif %}
        )
    """

    caption: Optional[str] = None

    def to_text(self):
        image = self.children[0].to_text()
        caption = []
        for idx, child in enumerate(self.children):
            if idx == 0:
                continue
            caption.append(child.to_text())
        return self.get_template().render(image=image, caption=caption)


class Link(Element):
    """Link to external website.

    :ref: https://typst.app/docs/reference/model/link/
    """

    LABEL = "link"
    TEMPLATE = """\
        #link(
          "{{ dest }}",
          [
            {{ content|indent(4, first=False) }}
          ],
        )
    """

    def __init__(
        self,
        uri: str,
        content: Optional[str] = None,
        parent=None,
        children=None,
        **kwargs,
    ):
        super().__init__(parent, children, **kwargs)
        self.uri = uri
        self.content = content or uri

    def to_text(self):
        return self.get_template().render(dest=self.uri, content=self.content)



"""Writer and relative classes."""
# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107

from __future__ import annotations

from typing import TYPE_CHECKING

from docutils import nodes
from sphinx.util.docutils import SphinxTranslator
from sphinx.util.logging import getLogger

from . import elements

if TYPE_CHECKING:
    from typing import Optional

    from sphinx.builders import Builder


logger = getLogger(__name__)


class TypstTranslator(SphinxTranslator):
    """Custom translator that has converter from dotctree to Typst syntax."""

    optional = [
        # TODO: Implement after for configure document itself.
        "document",
        # NOTE: Currently, these need not render anything.
        "description",
        "list_item",
        "field",
        "option_list_item",
        "option_group",
        "option",
        "colspec",
        "thead",
        "tbody",
        "row",
        "entry",
        "compound",
        "start_of_file",
    ]

    ELEMENT_MAPPING: dict[str, type[nodes.Element]] = {
        "paragraph": elements.Paragraph,
        "bullet_list": elements.BulletList,
        "field_list": elements.Element,
        "field": elements.Field,
        "field_name": elements.Element,
        "field_body": elements.Element,
        "option_list": elements.Table,
        "option_string": elements.Strong,
        "docinfo": elements.Element,
        "enumerated_list": elements.NumberedList,
        "emphasis": elements.Emphasis,
        "strong": elements.Strong,
        "block_quote": elements.Quote,
        "figure": elements.Figure,
    }
    """Controls for mapping Typst elements and docutils nodes.

    If a node requires only to add element with empty content,
    you should add pair of node name and element class into this.
    """

    def __init__(self, document: nodes.document, builder: Builder) -> None:
        super().__init__(document, builder)
        self.dom: elements.Document = elements.Document()
        self._ptr = self.dom
        # Set to avoid rendering root hedering text.
        self._section_level = -1
        self._targets: list[nodes.target] = []

    def _find_mepped_element(self, node) -> Optional[type[nodes.Element]]:
        """Find registed element mapped from node.

        :returns: Matched Element subclass. If it does not matche, return None.
        """
        for node_class in node.__class__.__mro__:
            if node_class.__name__ in self.ELEMENT_MAPPING:
                return self.ELEMENT_MAPPING[node_class.__name__]
        return None

    def _move_ptr_to_parent(self, node=None):
        """Move pointer for parent of current node.

        This method should be used to assign departure methods
        when only visit method require to define.
        """
        self._ptr = self._ptr.parent

    def unknown_visit(self, node: nodes.Node):
        """Handle visit methods for unregistered nodes.

        When node class has mapping for element class,
        it creates element object as child of current pointed node,
        and it moves pointer for created object.
        """
        element_class = self._find_mepped_element(node)
        if element_class is None:
            super().unknown_visit(node)
            return
        self._ptr = element_class(parent=self._ptr)

    def unknown_departure(self, node: nodes.Node):
        """Handle depart methods for unregistered nodes.

        When node class has mapping for element class,
        and it moves pointer for parent of current node.
        """
        element_class = self._find_mepped_element(node)
        if element_class is None:
            super().unknown_departure(node)
            return
        self._move_ptr_to_parent()

    def visit_raw(self, node: nodes.raw):
        if node.get("format") == "typst":
            elements.Source(node.astext(), parent=self._ptr)
        raise nodes.SkipNode()

    def visit_literal(self, node: nodes.raw):
        elements.Raw(node.astext(), parent=self._ptr)
        raise nodes.SkipNode()

    def visit_literal_block(self, node: nodes.raw):
        elements.RawBlock(node.astext(), node.get("language", None), parent=self._ptr)
        raise nodes.SkipNode()

    def visit_Text(self, node: nodes.Text):
        """Work about visit text content of node.

        This type should manage content value itself.
        """
        self._ptr = elements.Text(node.astext(), parent=self._ptr)

    depart_Text = _move_ptr_to_parent

    def visit_attribution(self, node: nodes.attribution):
        if isinstance(node.parent, nodes.block_quote):
            self._ptr.attribution = node.astext()
        raise nodes.SkipNode()

    def visit_Admonition(self, node):
        msg = "Currently, admonition-like directive is not supported."
        logger.info(msg)
        raise nodes.SkipNode()

    def visit_image(self, node: nodes.image):
        elements.Image(
            node["uri"], node.get("width"), node.get("alt"), parent=self._ptr
        )
        raise nodes.SkipNode()

    def visit_caption(self, node: nodes.caption):
        if isinstance(self._ptr, elements.Figure):
            para = elements.Paragraph(parent=self._ptr)
            elements.Text(node.astext(), parent=para)
        raise nodes.SkipNode()

    def visit_legend(self, node: nodes.legend):
        pass

    def depart_legend(self, node: nodes.legend):
        pass

    def visit_table(self, node: nodes.table):
        self._ptr = elements.Table(parent=self._ptr)

    depart_table = _move_ptr_to_parent

    def visit_tgroup(self, node: nodes.tgroup):
        self._ptr.columns = int(node["cols"])

    def depart_tgroup(self, node):
        pass

    def visit_reference(self, node: nodes.reference):
        self._ptr = elements.Link(node["refuri"], node.astext(), parent=self._ptr)

    depart_reference = _move_ptr_to_parent

    def visit_target(self, node: nodes.target):
        node_idx = node.parent.children.index(node)  # type: ignore[possibily-unbound-attribute]
        if node_idx < 0:
            self._targets.append(node)
        elif isinstance(self._ptr.children[-1], elements.Heading):
            self._ptr.children[-1].label = node["refid"]
        raise nodes.SkipNode()

    def visit_title(self, node: nodes.title):
        self._ptr = elements.Heading(parent=self._ptr)
        self._ptr.level = self._section_level
        if self._targets:
            target = self._targets.pop()
            self._ptr.label = target["refid"]

    depart_title = _move_ptr_to_parent

    def visit_section(self, node: nodes.section):
        self._ptr = elements.Section(parent=self._ptr)
        self._section_level += 1

    def depart_section(self, node):
        self._move_ptr_to_parent()
        self._section_level -= 1



"""Configuration for demo build that works correctly on current version."""

# -- Project information
project = "Demo of atsphinx-typst"
copyright = "2025, Kazuya Takei"
author = "Kazuya Takei"
release = "0.0.0"

# -- General configuration
extensions = []
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output
html_theme = "alabaster"
html_static_path = ["_static"]



# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107
import textwrap

from atsphinx.typst.core import elements as t


class TestSource:
    def test_single_line(self):
        elm = t.Source("$ A = pi r^2 $")
        assert elm.to_text() == "$ A = pi r^2 $\n"

    def test_multi_lines(self):
        elm = t.Source(
            textwrap.dedent("""\
            #layout(size => {
              let half = 50% * size.width
              [Half a page is #half wide.]
            })
        """)
        )
        assert elm.to_text() == textwrap.dedent("""\
            #layout(size => {
              let half = 50% * size.width
              [Half a page is #half wide.]
            })

        """)


class TestText:
    def test_single_line(self):
        elm = t.Text("Lorem ipsum dolor sit amet,")
        assert elm.to_text() == "Lorem ipsum dolor sit amet,"

    def test_multi_lines(self):
        elm = t.Text(
            textwrap.dedent("""\
            Lorem ipsum dolor sit amet, consectetur adipiscing elit,
            sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        """)
        )
        assert elm.to_text() == textwrap.dedent("""\
            Lorem ipsum dolor sit amet, consectetur adipiscing elit,
            sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        """)



"""Standard tests."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx("html")
def test__compatibility(app: SphinxTestApp):
    """Test to pass."""
    app.build()


@pytest.mark.sphinx("typst")
def test__it(app: SphinxTestApp):
    """Test to pass."""
    app.build()
    out = app.outdir / "index.typ"
    assert out.exists()
    assert "Test doc for atsphinx-typst" not in out.read_text()


@pytest.mark.sphinx("typst", confoverrides={"extensions": []})
def test__auto_adding_extension(app: SphinxTestApp):
    """Test to pass."""
    app.build()
    assert (app.outdir / "index.typ").exists()


@pytest.mark.sphinx(
    "typst",
    confoverrides={
        "typst_documents": [
            {
                "entry": "index",
                "filename": "document-1",
                "title": "test",
            },
            {
                "entry": "index",
                "filename": "document-2",
                "title": "test",
            },
        ]
    },
)
def test__multiple_output(app: SphinxTestApp):
    """Test to pass."""
    app.build()
    assert (app.outdir / "document-1.typ").exists()
    assert (app.outdir / "document-2.typ").exists()



"""Base and essential elements."""
# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107

from __future__ import annotations

import textwrap
from functools import lru_cache
from typing import TYPE_CHECKING

import jinja2
from anytree import Node

if TYPE_CHECKING:
    from typing import ClassVar


class Element(Node):
    """Abstract of all elements.

    This defines common methods and class vars.
    """

    LABEL: ClassVar[str] = ""
    """Name to show as anytree node."""
    TEMPLATE: ClassVar[str] = """\
        {%- for content in contents -%}
        {{ content }}
        {%- endfor -%}
    """
    """Template string when ``to_text`` runs."""

    def __init__(self, parent=None, children=None, **kwargs):
        """Set ``cls.LABEL`` for node name of anytree when it is created."""
        super().__init__(self.LABEL, parent, children, **kwargs)

    @classmethod
    @lru_cache()
    def get_template(cls) -> jinja2.Template:
        """Create template object from class vars."""
        return jinja2.Template(textwrap.dedent(cls.TEMPLATE).lstrip("\n"))

    def to_text(self):
        """Convert from element to Typst source."""
        return self.get_template().render(contents=[c.to_text() for c in self.children])


class Source(Element):
    """Raw Typst source (It is not Typst's ``raw`` element!).

    This is from Sphinx ``raw`` directive, and it is used to set Typst customize.
    """

    LABEL = "#raw"

    content: str
    """Content to insert into source."""

    def __init__(self, content: str, parent=None, children=None, **kwargs):
        super().__init__(parent, children, **kwargs)
        self.content = content

    def to_text(self):
        return self.content + "\n"


class Text(Element):
    """Plain text element."""

    LABEL = "#text"

    content: str
    """Text content itself."""

    def __init__(self, content: str, parent=None, children=None, **kwargs):
        super().__init__(parent, children, **kwargs)
        self.content = content

    def to_text(self):
        return self.content



"""Generate Typst sources and PDF from Sphinx document."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sphinx.application import Sphinx

__version__ = "0.0.0"


def setup(app: Sphinx):  # noqa: D103
    # Load sub extensions
    app.setup_extension("atsphinx.typst.core")
    return {
        "version": __version__,
        "env_version": 1,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Drawing elements.

:ref: https://typst.app/docs/reference/visualize/
"""
# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107

from __future__ import annotations

from typing import TYPE_CHECKING

from .base import Element

if TYPE_CHECKING:
    from typing import Optional


class Image(Element):
    """Embedding image.

    :ref: https://typst.app/docs/reference/visualize/image/
    """

    LABEL = "image"
    TEMPLATE = """\
        #image(
          "{{ elm.uri }}",
          {%- if elm.width %}
          width: {{ elm.width }},
          {%- endif %}
          {%- if elm.alt %}
          alt: "{{ elm.alt }}",
          {%- endif %}
        )
    """

    uri: str
    width: Optional[str]
    alt: Optional[str]

    def __init__(
        self,
        uri: str,
        width: Optional[str] = None,
        alt: Optional[str] = None,
        parent=None,
        children=None,
        **kwargs,
    ):
        super().__init__(parent, children, **kwargs)
        self.uri = uri
        self.width = width
        self.alt = alt

    def to_text(self):
        return self.get_template().render(elm=self)



# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107
import textwrap

from atsphinx.typst.core.elements import visualize as t


class TestImage:
    def test_default(self):
        elm = t.Image("./example.png")
        assert elm.to_text() == textwrap.dedent("""\
#image(
  "./example.png",
)""")

    def test_with_param(self):
        elm = t.Image("./example.png", alt="Dummy")
        assert elm.to_text() == textwrap.dedent("""\
#image(
  "./example.png",
  alt: "Dummy",
)""")



"""Classes root of Typst elements.

This module is entrypoint of importing core elements (written in submodules).
"""
# ruff: noqa: F401

from .base import Element, Source, Text
from .model import (
    BulletList,
    Document,
    Field,
    Figure,
    Heading,
    Link,
    NumberedList,
    Paragraph,
    Quote,
    Section,
    Table,
)
from .text import Emphasis, Raw, RawBlock, Strong
from .visualize import Image



# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107
import textwrap

from atsphinx.typst.core.elements import text as t
from atsphinx.typst.core.elements.base import Text


class TestRaw:
    def test_it(self):
        elm = t.Raw("Content")
        assert elm.to_text() == textwrap.dedent("""\
#raw(
  "Content"
)""")

    def test_with_escaped(self):
        elm = t.Raw('print("テスト")')
        assert elm.to_text() == textwrap.dedent("""\
#raw(
  "print(\\"テスト\\")"
)""")


class TestRawBlock:
    def test_single_it(self):
        elm = t.RawBlock(
            textwrap.dedent("""\
                print("テスト")
                print("Hello")
            """).strip("\n"),
            "python",
        )
        assert elm.to_text() == textwrap.dedent("""\
```python
print(\"テスト\")
print(\"Hello\")
```
""")


class TestEmphasis:
    def test_it(self):
        elm = t.Emphasis()
        Text("Content", parent=elm)
        assert elm.to_text() == textwrap.dedent("""\
#emph[
  Content
]""")


class TestStrong:
    def test_it(self):
        elm = t.Strong()
        Text("Content", parent=elm)
        assert elm.to_text() == textwrap.dedent("""\
#strong[
  Content
]""")



"""Custom builders."""

from __future__ import annotations

from datetime import date
from pathlib import Path
from typing import TYPE_CHECKING

from docutils import nodes
from sphinx import addnodes
from sphinx._cli.util.colour import darkgreen
from sphinx.builders import Builder
from sphinx.errors import SphinxError
from sphinx.locale import _
from sphinx.util.nodes import inline_all_toctrees

from . import themes, writer

if TYPE_CHECKING:
    from docutils import nodes

    from .config import DocumentSettings


class TypstBuilder(Builder):
    """Custom builder to generate Typst source from doctree."""

    name = "typst"
    format = "typst"
    default_translator_class = writer.TypstTranslator

    def get_outdated_docs(self):  # noqa: D102
        return "all targets"

    def prepare_writing(self, docnames: set[str]) -> None:  # noqa: D102
        # TODO: Implement after if it needs
        pass

    def write_documents(self, docnames):  # noqa: D102
        for document_settings in self.config.typst_documents:
            self.write_doc(document_settings)

    def write_doc(self, document_settings: DocumentSettings):  # noqa: D102
        docname = document_settings["entry"]
        theme = themes.get_theme(document_settings["theme"])
        doctree = self.assemble_doctree(docname, document_settings["toctree_only"])
        visitor: writer.TypstTranslator = self.create_translator(doctree, self)  # type: ignore[assignment]
        doctree.walkabout(visitor)
        today_fmt = self.config.today_fmt or _("%b %d, %Y")
        context = themes.ThemeContext(
            title=document_settings["title"],
            config=self.config,
            date=date.today().strftime(today_fmt),
            body=visitor.dom.to_text(),
        )
        out = Path(self.app.outdir) / f"{document_settings['filename']}.typ"
        theme.write_doc(out, context)

    def assemble_doctree(self, docname: str, toctree_only: bool) -> nodes.document:
        """Find toctree and merge children doctree into parent doctree.

        This method is to generate single Typst document.

        .. todo::

           We must see how does inline_all_toctrees work.
        """
        root = self.env.get_doctree(docname)
        if toctree_only:
            root_section = nodes.section()
            for toctree in root.findall(addnodes.toctree):
                root_section += toctree
            root = root.copy()
            root += root_section
        tree = inline_all_toctrees(self, {docname}, docname, root, darkgreen, [docname])
        return tree

    def get_target_uri(self, docname, typ=None):  # noqa: D102
        # TODO: Implement it!
        return ""


class TypstPDFBuilder(TypstBuilder):
    """PDF creation builder from doctree.

    This is similar to the relationship between
    the latexpdf builder and the latex builder.
    """

    name = "typstpdf"
    format = "typst"

    def init(self) -> None:
        """Check that python env has typst package."""
        try:
            import typst  # noqa - Only try importing
        except ImportError:
            raise SphinxError("Require 'typst' to run 'typstpdf' builder.")

    def write_doc(self, document_settings: DocumentSettings) -> None:  # noqa: D102
        # TODO: Implement it!
        import typst

        super().write_doc(document_settings)
        src = Path(self.app.outdir) / f"{document_settings['filename']}.typ"
        out = Path(self.app.outdir) / f"{document_settings['filename']}.pdf"
        typst.compile(src, output=out)



# TODO: Write more tests.
# ruff: noqa: D101, D102, D107
import textwrap

import pytest

from atsphinx.typst.core.elements import model as t
from atsphinx.typst.core.elements.base import Text


class TestHeading:
    def test_valid_level(self):
        elm = t.Heading()
        elm.level = 1
        Text("Section title", parent=elm)
        assert elm.to_text() == textwrap.dedent("""\
#heading(
  level: 1,
  [
    Section title
  ]
)""")

    def test_valid_level_2(self):
        elm = t.Heading()
        elm.level = 2
        Text("Section title", parent=elm)
        assert elm.to_text() == textwrap.dedent("""\
#heading(
  level: 2,
  [
    Section title
  ]
)""")

    def test_invalid_level(self):
        elm = t.Heading()
        elm.level = 0
        assert elm.to_text() == ""


class TestBulletList:
    @pytest.mark.skip(reason="Thinking cases")
    def test_it(self):
        pass


class TestNumberedList:
    @pytest.mark.skip(reason="Thinking cases")
    def test_it(self):
        pass


class TestTable:
    @pytest.mark.skip(reason="Thinking cases")
    def test_it(self):
        pass


class TestQuote:
    @pytest.mark.skip(reason="Thinking cases")
    def test_it(self):
        pass


class TestFigure:
    @pytest.mark.skip(reason="Thinking cases")
    def test_it(self):
        pass


class TestLink:
    def test_display_url(self):
        elm = t.Link("http://example.com")
        assert elm.to_text() == textwrap.dedent("""\
#link(
  "http://example.com",
  [
    http://example.com
  ],
)""")

    def test_with_text(self):
        elm = t.Link("http://example.com", content="EXAMPLE.COM")
        assert elm.to_text() == textwrap.dedent("""\
#link(
  "http://example.com",
  [
    EXAMPLE.COM
  ],
)""")



# noqa: D100

extensions = []

typst_documents = [
    {
        "entry": "index",
        "filename": "index",
        "theme": "manual",
        "title": "Test documentation",
        "toctree_only": True,
    }
]



"""Writer and relative classes."""

from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

import pytest
from docutils.core import publish_doctree

from atsphinx.typst.core import writer as t

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx("typst")
@pytest.mark.parametrize(
    "src,dest",
    [
        pytest.param(
            """
Paragraph.
""",
            """
Paragraph.
""",
            id="Single paragraph",
        ),
        pytest.param(
            """
This is paragraph 1.
That is too.
    """,
            """
This is paragraph 1.
That is too.
    """,
            id="Multiline paragraph",
        ),
        pytest.param(
            """
This is paragraph 1.

That is paragraph 2.
    """,
            """
This is paragraph 1.

That is paragraph 2.
    """,
            id="Multiple paragraph",
        ),
        pytest.param(
            """
Title
=====
    """,
            """
#heading(
  level: 1,
  [
    Title
  ]
)
    """,
            id="Single heading",
        ),
        pytest.param(
            """
Title
=====

Paragraph

Section 1
---------
    """,
            """
#heading(
  level: 1,
  [
    Title
  ]
)

Paragraph

#heading(
  level: 2,
  [
    Section 1
  ]
)
    """,
            id="Heading with paragraph",
        ),
        pytest.param(
            """
* Item A
* Item B
    """,
            """
#list(
  [
    Item A
  ],
  [
    Item B
  ]
)
    """,
            id="Bullet list",
        ),
        pytest.param(
            """
* Item A
  Next line
* Item B
    """,
            r"""
#list(
  [
    Item A
    Next line
  ],
  [
    Item B
  ]
)
    """,
            id="Bullet list with line break",
        ),
        pytest.param(
            """
* Item A

  * Sub item A
  * Sub item B

* Item B
    """,
            """
#list(
  [
    Item A
  ]
  +list(
    [
      Sub item A
    ],
    [
      Sub item B
    ]
  ),
  [
    Item B
  ]
)
    """,
            id="Bullet list with nested",
        ),
        pytest.param(
            """
* Item A

  #. Sub item A
  #. Sub item B

* Item B
    """,
            """
#list(
  [
    Item A
  ]
  +enum(
    [
      Sub item A
    ],
    [
      Sub item B
    ]
  ),
  [
    Item B
  ]
)
    """,
            id="Bullet list with nested numberd list",
        ),
        pytest.param(
            """
#. Item A
#. Item B
    """,
            """
#enum(
  [
    Item A
  ],
  [
    Item B
  ]
)
    """,
            id="Enumerated list",
        ),
        pytest.param(
            """
:Language: Japanese
:Language2: English
:Description: Hello world
              This is atsphinx-typst.
""",
            """
#sphinxField(
  [
    Language
  ],
  [
    Japanese
  ],
)#sphinxField(
  [
    Language2
  ],
  [
    English
  ],
)#sphinxField(
  [
    Description
  ],
  [
    Hello world
    This is atsphinx-typst.
  ],
)
    """,
            id="Docinfo",
        ),
        pytest.param(
            """
Paragraph

:Language: Japanese
""",
            """
Paragraph

#sphinxField(
  [
    Language
  ],
  [
    Japanese
  ],
)
    """,
            id="Field list",
        ),
        pytest.param(
            """
Paragraph

-h  Help
-v  Show version
""",
            """
Paragraph

#table(
  columns: 2,
  [
    #strong[
      -h
    ]
  ],
  [
    Help
  ],
  [
    #strong[
      -v
    ]
  ],
  [
    Show version
  ]
)
    """,
            id="Option list",
        ),
        pytest.param(
            """
*Content*
""",
            """
#emph[
  Content
]
    """,
            id="Emphasizes content",
        ),
        pytest.param(
            """
**Content**
""",
            """
#strong[
  Content
]
    """,
            id="strong content",
        ),
        pytest.param(
            """
This *is* **content**
""",
            """
This #emph[
  is
] #strong[
  content
]
    """,
            id="strong content",
        ),
        pytest.param(
            """
``print("テスト")``
""",
            """
#raw(
  "print(\\"テスト\\")"
)
""",
            id="inline raw code",
        ),
        pytest.param(
            """
This is an ordinary paragraph, introducing a block quote.

    "It is my business to know things.  That is my trade."
""",
            """
This is an ordinary paragraph, introducing a block quote.

#quote(
  block: true,
)[
  "It is my business to know things.  That is my trade."
]
""",
            id="block quote",
        ),
        pytest.param(
            """
This is an ordinary paragraph, introducing a block quote.

    "It is my business to know things.  That is my trade."

    -- Sherlock Holmes
""",
            """
This is an ordinary paragraph, introducing a block quote.

#quote(
  block: true,
  attribution: [Sherlock Holmes],
)[
  "It is my business to know things.  That is my trade."
]
""",
            id="block quote with attribution",
        ),
        pytest.param(
            """
.. code:: python

    print("テスト")
    print("Hello")
""",
            """
```python
print(\"テスト\")
print(\"Hello\")
```
""",
            id="block raw code",
        ),
        pytest.param(
            """
.. hint:: Hello

.. warning:: world
""",
            """
""",
            id="Admonitions",
        ),
        pytest.param(
            """
.. image:: ./example.jpg
""",
            """
#image(
  "./example.jpg",
)
""",
            id="Simple image",
        ),
        pytest.param(
            """
.. image:: ./example.jpg
    :width: 50%
    :alt: Sample
""",
            """
#image(
  "./example.jpg",
  width: 50%,
  alt: "Sample",
)
""",
            id="Image with attributes",
        ),
        pytest.param(
            """
.. figure:: ./example.jpg
""",
            """
#figure(
  #image(
    "./example.jpg",
  ),
)
""",
            id="Simple figure",
        ),
        pytest.param(
            """
.. figure:: ./example.jpg

    This is sample.
""",
            """
#figure(
  #image(
    "./example.jpg",
  ),
  caption: [
    This is sample.
  ],
)
""",
            id="Figure with caption",
        ),
        pytest.param(
            """
.. figure:: ./example.jpg

    This is sample.

    Support multiline.
""",
            # NOTE: It removes blank line between caption and legend.
            """
#figure(
  #image(
    "./example.jpg",
  ),
  caption: [
    This is sample.
    Support multiline.
  ],
)
""",
            id="Figure with caption",
        ),
    ],
)
def test_syntax(app: SphinxTestApp, src: str, dest: str):
    # NOTE: Keep debugging print
    from anytree import RenderTree

    """Very simple test for syntax by Translator."""
    document = publish_doctree(src.strip())
    print(document)
    document.settings.strict_visitor = False
    visitor = t.TypstTranslator(document, app.builder)
    visitor._section_level = 1
    document.walkabout(visitor)
    print(RenderTree(visitor.dom))
    assert visitor.dom.to_text().strip() == dest.strip()


@pytest.mark.parametrize(
    "src",
    [
        pytest.param(
            """
== ==== ========
ID Name Note
== ==== ========
1  Tom  Engineer
== ==== ========
""",
            id="Simple table",
        ),
        pytest.param(
            """
+--+----+--------+
|ID|Name|Note    |
+==+====+========+
|1 |Tom |Engineer|
+--+----+--------+
""",
            id="Grid table",
        ),
        pytest.param(
            """
.. csv-table::
    :header: ID,Name,Note

    1,Tom,Engineer
""",
            id="CSV table",
        ),
        pytest.param(
            """
.. list-table::

    * - ID
      - Name
      - Note
    * - 1
      - Tom
      - Engineer
""",
            id="List table",
        ),
    ],
)
def test_table_syntax(app: SphinxTestApp, src: str):
    dest = """
#table(
  columns: 3,
  [
    ID
  ],
  [
    Name
  ],
  [
    Note
  ],
  [
    1
  ],
  [
    Tom
  ],
  [
    Engineer
  ]
)
"""
    # NOTE: Keep debugging print
    from anytree import RenderTree

    """Very simple test for syntax by Translator."""
    document = publish_doctree(src.strip())
    print(document)
    document.settings.strict_visitor = False
    visitor = t.TypstTranslator(document, app.builder)
    document.walkabout(visitor)
    print(RenderTree(visitor.dom))
    assert visitor.dom.to_text().strip() == dest.strip()


@pytest.mark.parametrize(
    "src,dest",
    [
        pytest.param(
            """
http://example.com
""",
            """
#link(
  "http://example.com",
  [
    http://example.com
  ],
)
""",
        ),
        pytest.param(
            """
`Here <http://example.com>`_
""",
            """
#link(
  "http://example.com",
  [
    Here
  ],
)
""",
        ),
        pytest.param(
            """
`a link`_

.. _a link: http://example.com/
""",
            """
#link(
  "http://example.com/",
  [
    a link
  ],
)
""",
        ),
        pytest.param(
            """
.. _my-reference-label:

Section to cross-reference
--------------------------
""",
            """
#heading(
  level: 1,
  [
    Section to cross-reference
    #label("my-reference-label")
  ]
)
""",
        ),
    ],
)
def test_reference(app: SphinxTestApp, src: str, dest: str):
    # NOTE: Keep debugging print
    from anytree import RenderTree

    """Very simple test for syntax by Translator."""
    document = publish_doctree(src.strip())
    print(document)
    visitor = t.TypstTranslator(document, app.builder)
    visitor._section_level = 1
    document.walkabout(visitor)
    print(RenderTree(visitor.dom))
    assert visitor.dom.to_text().strip() == dest.strip()


def test_raw_typst_source(app: SphinxTestApp):
    src = """
        .. raw:: typst

            #heading([Hello])
        """
    document = publish_doctree(textwrap.dedent(src))
    visitor = t.TypstTranslator(document, app.builder)
    dest = """
        #heading([Hello])

        """
    document.walkabout(visitor)
    assert visitor.dom.to_text() == textwrap.dedent(dest)


def test_raw_not_typst_source(app: SphinxTestApp):
    src = """
        .. raw:: python

            print("hello")
        """
    document = publish_doctree(textwrap.dedent(src))
    visitor = t.TypstTranslator(document, app.builder)
    document.walkabout(visitor)
    assert visitor.dom.to_text() == textwrap.dedent("")



# noqa: D100

extensions = [
    "atsphinx.typst",
]

typst_documents = [
    {
        "entry": "index",
        "filename": "index",
        "theme": "manual",
        "title": "Test documentation",
    }
]



"""Text styling elements.

These elements can be created by reStructuredText's standard syntax.

:ref: https://typst.app/docs/reference/text/
"""
# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107

from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

from ._templating import env
from .base import Element

if TYPE_CHECKING:
    import jinja2


class Raw(Element):
    """Inline highlighting element.

    :ref: https://typst.app/docs/reference/text/raw/
    """

    TEMPLATE = """\
        #raw(
          {{ content|tojson|indent(2, first=False)}}
        )
    """

    content: str

    def __init__(self, content: str, parent=None, children=None, **kwargs):
        super().__init__(parent, children, **kwargs)
        self.content = content

    @classmethod
    @lru_cache()
    def get_template(cls) -> jinja2.Template:
        # Override to work 'ensure_ascii' settings by tojson.
        return env.get_template("Raw")

    def to_text(self):
        return self.get_template().render(content=self.content)


class RawBlock(Element):
    """Code-block element.

    Currently, this element doesn't render as function style
    to control easily from Sphinx.

    :ref: https://typst.app/docs/reference/text/raw/
    """

    TEMPLATE = """\
        ```{{lang}}
        {{content}}
        ```

    """

    content: str
    lang: str
    """Highlighting language."""

    def __init__(self, content: str, lang: str, parent=None, children=None, **kwargs):
        super().__init__(parent, children, **kwargs)
        self.content = content
        self.lang = lang

    def to_text(self):
        return self.get_template().render(content=self.content, lang=self.lang)


class FunctionalText(Element):
    """Element base-class to render decorated text.

    This is for inline syntax.
    """

    TEMPLATE = """\
        #{{label}}[
          {%- for content in contents %}
          {{ content | indent(2, first=False) }}
          {%- endfor %}
        ]
    """

    def to_text(self):
        return self.get_template().render(
            label=self.LABEL, contents=[c.to_text() for c in self.children]
        )


class Emphasis(FunctionalText):
    """Emphasized text.

    :ref: https://typst.app/docs/reference/model/emph/
    """

    LABEL = "emph"


class Strong(FunctionalText):
    """Strong emphasized text.

    :ref: https://typst.app/docs/reference/model/strong/
    """

    LABEL = "strong"



"""Templating environment of core elements."""
# TODO: Write docstrings after.
# ruff: noqa: D101, D102, D107

from __future__ import annotations

import sys
import textwrap

import jinja2


def load_template(name: str) -> str:
    """Load template from class var.

    :param name: Class name to use as template.
    :returns: Template string.
    """
    import atsphinx.typst.core.elements  # noqa

    module = sys.modules["atsphinx.typst.core.elements"]
    if not hasattr(module, name):
        raise Exception(f"{name} is not found.")
    return textwrap.dedent(getattr(module, name).TEMPLATE).strip("\n")


env = jinja2.Environment(loader=jinja2.FunctionLoader(load_template))
env.policies["json.dumps_kwargs"]["ensure_ascii"] = False



"""Core features of Typst buider.

This only provides builder with features for Sphinx and doctree components.
This is registered standalone
when call builder without register into ``extensions`` of ``conf.py``.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .. import __version__
from . import builders, config

if TYPE_CHECKING:
    from sphinx.application import Sphinx


def setup(app: Sphinx):  # noqa: D103
    # Builders
    app.add_builder(builders.TypstBuilder)
    app.add_builder(builders.TypstPDFBuilder)
    config.setup(app)
    return {
        "version": __version__,
        "env_version": 1,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Cases for toctree option of document settingds."""

from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx("typst", testroot="toctree")
def test__it(app: SphinxTestApp):
    """Test to pass."""
    app.build()
    out = app.outdir / "index.typ"
    print(out.read_text())
    assert out.exists()
    assert "#heading(" in out.read_text()
    assert "Test doc for atsphinx-typst" not in out.read_text()
    assert (
        textwrap.dedent("""\
        #heading(
          level: 1,
          [
            Section title 1
          ]
        )""")
        in out.read_text()
    )
    assert (
        textwrap.dedent("""\
        #heading(
          level: 2,
          [
            Sub section 1-1
          ]
        )""")
        in out.read_text()
    )



"""Configuration."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from sphinx.application import Sphinx
    from sphinx.config import Config


class DocumentSettings(TypedDict):
    """Build settings each documets."""

    entry: str
    """Docname of entrypoint."""
    filename: str
    """Output filename (without ext)."""
    title: str
    """Title of document."""
    theme: str
    """Generate theme."""
    toctree_only: bool
    """When it is ``True``, builder only write contents of toctree from 'entry'."""


DEFAULT_DOCUMENT_SETTINGS = {
    "theme": "manual",
    "toctree_only": False,
}


def set_config_defaults(app: Sphinx, config: Config):
    """Inject default values of configured ``typst_documents``."""
    document_settings = config.typst_documents or []
    if not document_settings:
        document_settings.append(
            {
                "entry": config.root_doc,
                "filename": f"document-{config.language}",
                "title": f"{config.project} Documentation [{config.language.upper()}]",
                "theme": "manual",
            }
        )
    for idx, user_value in enumerate(document_settings):
        document_settings[idx] = DEFAULT_DOCUMENT_SETTINGS | user_value
    config.typst_documents = document_settings


def setup(app: Sphinx):  # noqa: D103
    app.add_config_value("typst_documents", [], "env", list[dict])
    app.connect("config-inited", set_config_defaults)



"""Theme management of Typst builder."""

from __future__ import annotations

from dataclasses import asdict, dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from jinja2 import Template

if TYPE_CHECKING:
    from sphinx.config import Config

_HERE = Path(__file__).parent


@dataclass
class Theme:
    """Document theme component."""

    template_path: Path

    def get_template(self) -> Template:
        """Retrieve template to render Typst source."""
        return Template(self.template_path.read_text(encoding="utf8"))

    def write_doc(self, out: Path, context: ThemeContext):
        """Write content as document."""
        tmpl = self.get_template()
        content = tmpl.render(asdict(context))
        out.parent.mkdir(parents=True, exist_ok=True)
        out.write_text(content, encoding="utf8")


@dataclass
class ThemeContext:
    """Default context values for templating."""

    title: str
    config: Config
    date: str
    body: str


def get_theme(name: str) -> Theme:
    """Find and setup built-in theme."""
    theme_path = _HERE / name
    return Theme(template_path=theme_path / "page.typ.jinja")
